# A simple script to use with a Statestep model treated
# as a finite state machine.
# A suitable model with which to try this is harelsWatch
# (one of the sample models that comes with Statestep) -
# if you have generated harelsWatch.pm, do:
#
#   perl -MharelsWatch=:all fsm.pl

BEGIN {
    die 'No Statestep model (run "perl -M<model>=:all fsm.pl")'
        unless defined $Combination;
    die 'Model contains no variables with values' unless @Vars;
    die 'Model contains no events' unless @Events;
}

use 5.008008;
use warnings;
use strict;

# map event identifiers to numbers
my %event_num = map { $Events[$_] => $_ } 0 .. $#Events;

{
    no strict 'refs';
    no warnings 'redefine';

    # redefine the 'str' method of the Combination class
    local *{ref($Combination->new()) . '::str'} = sub {
            my $state = $_[0];
            join "\n", map sprintf("%20s = %s", $Vars[$_],
                $Values[$_][$state->value($_)]), 0 .. $#Vars;
        };

    fsm();
}

# basic finite state machine
sub fsm {
    # try to find a valid initial state
    my $state = $Combination->iter_valid()->next();
    $state = $Combination->new() unless defined $state;
    print "Initial state:\n";
    for (;;) {
        print $state->str(), "\n";
        print '(This state is combination ', $state->num(), ".)\n";
        unless ($state->valid()) {
            my $expl = $state->explain_invalid();
            if ($expl =~ /^\d/) {
                $expl = "'$Constraints[-1][$expl]'";
            }
            print 'Note: this state is not valid (according to ',
                "constraint\n$expl).\n";
        }
        # user-defined subroutine
        my $num = get_next_event();
        if (!$state->expected($num)) {
            my $con = $state->explain_unexpected($num);
            print 'Note: this event was not expected ',
                "in this state\n(see constraint '",
                $Constraints[$num][$con],
                "' for event '$Events[$num]').\n";
        }
        if (my ($rule, $next) = $state->apply_rules($num)) {
            print "Rule '$Rules[$num][$rule]' for ",
                "event '$Events[$num]' applies. New state:\n";
            $state = $next;
        }
        else {
            print "No rule applies for event '$Events[$num]' ",
                "in this state.\n";
        }
    }
}

# read an event from standard input
sub get_next_event {
    print "Enter an event ('.' to quit):\n";
    for (;;) {
        print '> ';
        my $in = <>;
        defined $in or print("\n"), exit 0;
        print $in unless -t;
        if (my ($id) = $in =~ /(\S+)/) {
            if (exists $event_num{$id}) {
                return $event_num{$id};
            }
            elsif ($id eq '.') {
                exit 0;
            }
            else {
                print '? The events are: ',
                    join(', ', @Events), ".\n";
            }
        }
    }
}
