<#ftl strip_whitespace = true>
<#-- Variables without any values must be ignored. -->
<#if doc.model.variables?size != 0>
    <#assign usable_variables =
    doc.model.variables["variable[./value]"]
    />
<#else>
    <#assign usable_variables = []/>
</#if>
<#assign outPath = pp.realOutput?replace('.pm$', '', 'r')/>
<#assign outPath = outPath?split(pp.slash)?reverse/>
<#assign package = outPath[0]/>
<#if ((name_dirs!0) > 0)>
    <#list 1 .. name_dirs as x>
        <#assign package = outPath[x] + "::" + package/>
    </#list>
</#if>

=head1 NAME

${package} - generated from Statestep template '${pp.sourceFileName}'

=head1 SYNOPSIS

Autogenerated code which implements the logic
of a Statestep model.

    use ${package} ':all';

=head1 DESCRIPTION

=cut

package ${package};

use 5.008008;
use warnings;
use strict;
use Math::BigInt;

BEGIN {
    use Exporter   ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION     = 0.01;
    @ISA         = qw(Exporter);
    @EXPORT      = qw();
    %EXPORT_TAGS = (
        all => [qw(@Vars @Values
            $Combination $Combinations
            @Events
            @Constraints @Constraint_Texts
            @Rules @Rule_Texts)]);
    Exporter::export_ok_tags(keys %EXPORT_TAGS);
}

=head2 ${'C<@Vars>'}

The names of the variables in the model.

=cut

our @Vars = (
<#list usable_variables as var>
    '${var.@name}'<#rt>
    <#if var_has_next>
        ,<#lt>
    </#if>
</#list>
);

our %_Var_Indices = map { $Vars[$_] => $_ } 0 .. $#Vars;

=head2 ${'C<@Values>'}

For each variable, the names of that variable's values.

    print "The first value of the variable '$Vars[0]'",
        " is '$Values[0][0]'.\n";

=cut

our @Values = (
<#list usable_variables as var>
    # ${var.@name}
    ['<#rt>
    <#list var[".//value[not(./value)]"] as leaf>
        <#list leaf?ancestors('value')?reverse as branch>
            ${branch.@name}-<#t>
        </#list>
        ${leaf.@name}'<#t>
        <#if leaf_has_next>
            ,<#lt>
    '<#rt>
        </#if>
    </#list>
    ]<#t>
    <#if var_has_next>
        ,<#lt>
    </#if>
</#list>
);

our @_Val_Offsets;
for my $v (0 .. $#Vars) {
    %{$_Val_Offsets[$v]} =
    map { $Values[$v][$_] => $_ } 0 .. $#<#rt>
    {$Values[$v]};<#lt>
}

=head2 C<$Combination>

This is simply 'L</${package}::Combination>' -
the name of the Combination class incorporated within
this package.

=cut

our $Combination = '${package}::Combination';

my $_Vals = '${package}::_Vals';
my $_CombinationMask = '${package}::_CombinationMask';
my $_TargetRow = '${package}::_TargetRow';
my $_CombIter = '${package}::_CombIter';
my $_AllowedByTableIter = '${package}::_AllowedByTableIter';
my $_AllowedByFormulasIter = '${package}::_AllowedByFormulasIter';
my $_GangIter = '${package}::_GangIter';

=head2 C<$Combinations>

The number of distinct Combinations.
(This is currently a C<Math::BigInt>, but should be
treated as a regular integer.)

=cut

our $Combinations = Math::BigInt->bone();
$Combinations *= @$_ for @Values;

=head2 ${'C<@Events>'}

The event names.

    print "The first event is '$Events[0]'.\n" if @Events;

=cut

our @Events = (<#t>
<#list doc.model.events.event as event>

    '${event.@name}'<#rt>
    <#if event_has_next>
        ,<#t>
    </#if>
</#list>
);

# constraining values are first in each row
our @_Constraint_Table = (<#t>
<#assign have_row = false/>
<#list doc.model.simple_constraints.constraint_group as row>
    <#if (row["./value_set[./@type='constrained']
    [not(./@content='dont_care')]"] ?size > 0)>
        <#if have_row>
            ,<#t>
        </#if>
        <#assign have_row = true/>

    [${values_to_code(row[
    "./value_set[./@type='constraining']"][0])},<#rt>
        <#list row["./value_set[./@type='constrained']
        [not(./@content='dont_care')]"] as vals>

        ${values_to_code(vals)}<#rt>
            <#if vals_has_next>
                , <#t>
            <#else>
                ]<#t>
            </#if>
        </#list>
    </#if>
</#list>
);

use constant {
<#list ["SOURCE", "TARGET", "NOT", "AND", "OR", "IMPLIES"] as x>
    _${x} => ${x_index}<#rt>
    <#if x_has_next>
        ,<#lt>
    </#if>
</#list>
};

=head2 ${'C<@Constraints>'}

Constraint identifiers indexed by event number
and constraint number.
For global constraints, use C<-1> for the event number.

    if (@Constraints[-1]) {
        print "The first global constraint is '" .
            $Constraints[-1][0] . "'.\n";
    }

=head2 ${'C<@Constraint_Texts>'}

The text of constraint formulas indexed by event number
and constraint number.
For global constraints, use C<-1> for the event number.

    if (@{$Constraint_Texts[0]}) {
        print "The formula of the first constraint for event '",
            "$Events[0]' is '$Constraint_Texts[0][0]'.\n";
    }

=cut

our @Constraints;
our @Constraint_Texts;
our @_Constraint_Formulas;
our @_Constraint_Vars;
# global constraints:
<@constraint_formulas doc.model "@Events"/>
<#list doc.model.events.event as event>
# '${event.@name}' constraints:
    <@constraint_formulas event event_index/>
</#list>

=head2 ${'C<@Rules>'}

Rule identifiers indexed by event number and rule number.

    if (my ($rule, $new_state) = $state->apply_rules($event)) {
        print "Rule [$Events[$event]] " .
            "$Rules[$event][$rule] applied.\n";
    }

=head2 ${'C<@Rule_Texts>'}

The text of rule formulas indexed by event number and rule number.
(May be useful where information has been embedded as comments in
the model's rule formulas.)

=cut

our @Rules;
our @Rule_Texts;
our @_Rules_Table;
our @_Rule_Formulas;
<#list doc.model.events.event as event>
# event '${event.@name}'
    <#if event.rules.rule?size == 0>
        $Rules[${event_index}] = [];<#lt>
        $_Rules_Table[${event_index}] = [];<#lt>
        $Rule_Texts[${event_index}] = [];<#lt>
    </#if>
    <#list event.rules.rule as rule>
$Rules[${event_index}][${rule_index}] = '${rule.@name}';
$_Rules_Table[${event_index}][${rule_index}] = [
    $_CombinationMask->new(<#rt>
        <#list rule[
        "./values_array/value_set[@type='source']
        [not(@content='dont_care')]"]
        as source_vals
        >

        ${values_to_code(source_vals)}<#rt>
            <#if source_vals_has_next>
            ,<#t>
            </#if>
        </#list>
),
    $_TargetRow->new(<#rt>
        <#list targetvals(rule) as arg>

        ${arg}<#rt>
            <#if arg_has_next>
            ,<#t>
            </#if>
        </#list>
)];
        $Rule_Texts[${event_index}][${rule_index}] = <#t>
        <#if rule.formula.text?size != 0>
            ${perl_string(rule.formula.text)};<#lt>
        <#else>
            '';<#lt>
        </#if>
        <#if rule.formula.parse_node?size != 0>
            $_Rule_Formulas[${event_index}][${rule_index}] =<#t>
            <@parse_node_to_code node=rule.formula.parse_node/>
;
        </#if>
    </#list>
</#list>

# A set of values of a variable.
{
    package ${package}::_Vals;
    use warnings;
    use strict;

    sub new {
        my $invocant = shift;
        my $var_name = shift;
        my $var = $_Var_Indices{$var_name};
        my $all_vals = $Values[$var];
        # assume the number of bits in an integer will
        # be no less than the maximum number of values
        # in a Statestep variable (30)
        my $vals = 0;
        for my $arg (@_) {
            (my $re = $arg) =~ s/\*/.*/;
            $re = qr/^$re$/;
            my $val = 0;
            my $bit = 1;
            for (@$all_vals) {
                $val |= $bit if /$re/;
                $bit <<= 1;
            }
            $val or die "'$arg' is not a value of '$var_name'";
            $vals |= $val;
        }
        my $self = [$var, $vals];
        return bless $self, ref($invocant) || $invocant;
    }
    sub str {
        my $vals = shift;
        my ($var, $bits) = @$vals;
        my @valstrs;
        my $bit = 1;
        for (@{$Values[$var]}) {
            push @valstrs, $_ if $bit & $bits;
            $bit <<= 1;
        }
        return $Vars[$var] . '={' . join('|', @valstrs) . '}';
    }
    sub var_index {
        my $self = shift;
        return $$self[0];
    }
    sub val_bits {
        my $self = shift;
        return $$self[1];
    }
}

=head2 C<${package}::Combination>

An instance of this class maps each variable
to one of its values.

For convenience, L</$Combination>
holds the name of this package:

    my $comb = $Combination->new();

Distinct Combinations are ordered.
In the first Combination (numbered 0),
each variable C<$Vars[x]> has the value C<$Values[x][0]> -
this is the variable's default value in a Combination
(its value unless specified in the constructor).
The comparison operators are overloaded:

    my $prev;
    for my $comb (sort @combs) {
        # skip duplicate Combinations in @combs
        next if defined($prev) && $comb == $prev;
        $prev = $comb;
        # ...
    }

Like other references, Combinations evaluate as true in
a boolean context, and all Combinations are "greater than"
C<undef> (though comparisons generate the usual warning).

Efficient iterators are included which may be useful where
this code is used to test other software.
Naturally, however, these only increase the limit on how
big a model can be before exhaustive tests of the kind
supported by the iterators begin to take an impractically
long time to run.

=cut

{
    package ${package}::Combination;
    use warnings;
    use strict;
    use Carp qw/croak/;
    use overload '<=>' => \&_cmp, 'cmp' => \&_cmp, '""' => \&str;

    my $_default = $Combination->num(0);

    my $_instance_err_msg = ' not called as an instance method';

=head3 C<new()>

This constructor accepts a list of variable names and
value names in pairs.
If it is called as a class method then any unspecified
variables are given their default values.

    # these are equivalent: both create Combination 0,
    # in which all variables have their default values
    $c0 = $Combination->new();
    $c0 = $Combination->new($Vars[0], $Values[0][0]);

If it is called on an existing instance then it creates
a modified copy of that Combination, changing only the
specified values.

    $x = $Combination->new('VarTwo', 'two_B');
    $y = $Combination->new(qw/VarOne one_C VarTwo two_B/);
    $z = $x->new(VarOne=>'one_C');
    # now $z == $y

(The only way to "change" a Combination is to replace it
with a modified copy: Combinations are immutable by design.)

=cut

    sub new {
        my $invocant = shift;
        my @self = ref($invocant) ? @$invocant : @$_default;
        while (@_) {
            my $var_name = shift;
            my $var = $_Var_Indices{$var_name};
            croak "new(): no such variable '$var_name'"
                unless defined $var;
            my $val_name = shift;
            croak "new(): no value specified for '$var_name'"
                unless defined $val_name;
            my $offset = $_Val_Offsets[$var]{$val_name};
            croak "new(): '$val_name' is not a value of '$var_name'"
                unless defined $offset;
            $self[$var] = $offset;
        }
        return bless \@self, ref($invocant) || $invocant;
    }

=head3 C<num()>

When called as a class method, C<num()> is a constructor.

    # 0 <= $number < $Combinations
    my $state = $Combination->num($number);

When called on an instance, it returns the number of
the Combination (as a C<Math::BigInt>).

    print $Combination->new(3)->num(), "\n";
    # outputs '3' (assuming 3 < $Combinations)

This method is slow.
It is generally possible to avoid it by using
L<iterators|/iter()> and by comparing Combinations
directly using the overloaded comparison operators
(C<==>, C<E<lt>>, etc.).

=cut

    sub num {
        if (ref $_[0]) {
            my $self = shift;
            @_ and croak 'num() used as an instance method ' .
                'expects no arguments';
            my $n = Math::BigInt->bzero();
            my $factor = Math::BigInt->bone();
            for (reverse 0 .. $#Vars) {
                $n += $factor * $$self[$_];
                $factor *= @{$Values[$_]};
            }
            return $n;
        }
        else {
            my $class = shift;
            my $n = shift;
            defined($n) or croak 'num() used as a class method'
                . ' without an argument';
            my @self;
            eval {
                    no warnings;
                    $n =~ /^\d+$/ or die;
                    # for speed, treat Math::BigInt separately
                    if (ref $n) {
                        $n = $n->copy();
                        for (reverse 0 .. $#Vars) {
                            $self[$_] = ($n->bdiv(scalar
                                @{$Values[$_]}))[1]->numify();
                        }
                    }
                    else {
                        use integer;
                        for (reverse 0 .. $#Vars) {
                            my $num_vals = @{$Values[$_]};
                            $self[$_] = $n % $num_vals;
                            $n /= $num_vals;
                        }
                    }
                };
            $@ and croak 'num(): argument must be a '
                . 'positive integer or Math::BigInt';
            $n != 0 and croak 'num(): argument out of range';
            return bless \@self, $class;
        }
    }

    # $self->_cmp($arg) == ($self->num() <=> $arg->num())
    sub _cmp {
        if (!defined($_[1])) {
            warnings::warnif('uninitialized',
                'Use of uninitialized value' .
                ' (Combination compared with undef)');
            # for calls from module 'overload', $_[2] is 'swap'
            return $_[2] ? -1 : 1;
        }
        my $cmp = 0;
        eval {
                no warnings;
                $_[1]->_bits(0);
                for my $i (0 .. $#Vars) {
                    $cmp = $_[0][$i] <=> $_[1][$i];
                    return $cmp if $cmp;
                }
            };
        croak 'Combination compared with non-Combination' if $@;
        return $cmp;
    }

=head3 C<str()>

Returns a brief string representation of the Combination.
Through overloading, this method is automatically called
where a string is required.

    print $comb->str() . " == $comb\n";

=cut

    sub str {
        my $self = shift;
        ref $self or croak 'str()' . $_instance_err_msg;
        '[' . join(', ', map $self->value($_), @Vars) . ']';
    }

=head3 C<value()>

Given a variable, get its value in this Combination.
This method can be used in two modes: string or numeric.
If it is passed a variable name then it returns a value name:

    print "Variable '$Vars[0]' has value '" .
        $comb->value($Vars[0]) . "' in \$comb.\n";

If it is passed an index into ${'C<@Vars>'} then it returns
returns an index into the corresponding value names array:

    print "Variable '$Vars[0]' has value '",
        $Values[0][$comb->value(0)] . "' in \$comb.\n";

=cut

    sub value {
        my ($self, $arg) = @_;
        ref $self or croak 'value()' . $_instance_err_msg;
        defined($arg) or croak 'value() expects an argument';
        if ($arg =~ /^\d+$/) {
            $arg < @Vars or croak 'value(): index out of range';
            return $$self[$arg];
        }
        else {
            my $i = $_Var_Indices{$arg};
            defined($i) or croak "No such variable '$arg'";
            return $Values[$i][$$self[$i]];
        }
    }

    sub _bits {
        my ($self, $var_num) = @_;
        1 << $$self[$var_num];
    }

    # Arguments (optional): array of indices into @Vars.
    # Returns two subroutine references ($next, $reset).
    # $next() is an iterator subroutine which, starting with
    # the current Combination, returns Combinations in which
    # the values of the specified variables do not change;
    # the number of each successive Combination is the
    # smallest number greater than the previous one which
    # satisfies this property.
    # $reset() takes a Combination as argument, defaulting
    # to the default Combination. It resets the next
    # Combination returned by $next to the argument or to the
    # first Combination following it in which the variables
    # passed to _iter_subs() have the same values as they
    # had in the Combination on which _iter_subs() was
    # called.
    sub _iter_subs {
        my $comb = shift()->new();
        my @keep;
        $keep[$_] = 1 for @_;
        my $done = 0;
        return (
            sub {
                return if $done;
                my $next = $comb->new();
                my $i;
                for ($i = $#Vars; $i >= 0; --$i) {
                    next if $keep[$i];
                    last if (++$$comb[$i] < @{$Values[$i]});
                    $$comb[$i] = 0;
                }
                $done = $i < 0;
                return $next;
            },
            sub {
                my $new_comb = $_[0] || $_default;
                # index of most significant value that is greater
                # in the new iterator Combination than in the arg.
                my $incr = @Vars;
                for (my $i = $#Vars; $i >= 0; --$i) {
                    if ($keep[$i]) {
                        my $cmp = $$comb[$i] <=>
                            $$new_comb[$i] + ($incr == $i);
                        if ($cmp) {
                            $incr = $i;
                            --$incr if $cmp < 0;
                        }
                    }
                    else {
                        $$comb[$i] = $$new_comb[$i];
                        if ($i == $incr) {
                            next if ++$$comb[$i] < @{$Values[$i]};
                            $$comb[$i] = 0;
                            --$incr;
                        }
                    }
                }
                $done = $incr < 0;
                if (!$done) {
                    for ($incr + 1 .. $#Vars) {
                        $$comb[$_] = 0 unless $keep[$_];
                    }
                }
            });
    }

    # Arguments: indices into @Vars.
    # Returns an iterator subroutine which returns a series of
    # Combinations such that, when added to the Combination on
    # which this method was called:
    # 1. the values of the specified variables (taken together)
    #    are different in every Combination;
    # 2. The number of each Combination is the lowest number
    #    greater than the previous one which satisfies this
    #    property.
    # Thus, the first Combination returned by the iterator
    # $comb->_iter_change() is the first Combination following
    # $comb in which the values of the specified variables
    # are different from those in $comb.
    # (The algorithm is explained at the end of this file.)
    sub _iter_change {
        my $self = shift;
        my @indices = reverse sort @_;
        # index into @indices referencing the most significant
        # variable in that list which has been incremented so far
        my $imax = 0;
        # 1 + number of Combinations returned so far
        my $count = 1;
        # number of different possible combinations of values of
        # the variables referenced by @indices[$imax..0]
        my $cycles = @{$Values[$indices[$imax]]};
        my $comb = $self->new();
        # reset values of least significant unspecified variables
        # (minimizes the num() of the first Combination returned)
        $$comb[$_] = 0 for $indices[$imax] + 1 .. $#Vars;
        return sub {
            # most significant variable incremented (indexes @Vars)
            my $msvi;
            if ($count < $cycles) {
                for my $i (0 .. $imax) {
                    $msvi = $indices[$i];
                    last if (++$$comb[$msvi] < @{$Values[$msvi]});
                    $$comb[$msvi] = 0;
                }
            }
            else {
                until ($count < $cycles) {
                    return unless $imax < $#indices;
                    $msvi = $indices[++$imax];
                    $cycles *= @{$Values[$msvi]};
                }
                $$comb[$_] = 0 for $msvi + 1 .. $#Vars;
                unless (++$$comb[$msvi] < @{$Values[$msvi]}) {
                    $$comb[$msvi] = 0;
                }
            }
            unless ($$comb[$msvi]) {
                for (;;) {
                    if (!$msvi) {       # we've wrapped around
                        $cycles = 0;    # ensure future calls fail
                        return;
                    }
                    --$msvi;
                    last if (++$$comb[$msvi] < @{$Values[$msvi]});
                    $$comb[$msvi] = 0;
                }
                $cycles *= @{$Values[$indices[++$imax]]}
                    while $imax < $#indices
                    and $indices[$imax + 1] >= $msvi;
            }
            ++$count;
            return $comb->new();
        };
    }

=head3 C<valid()>

Returns true if this Combination is valid (that is,
if it is allowed by all of the global constraints),
otherwise false.

=cut

    sub valid {
        ref $_[0] or croak 'valid()' . $_instance_err_msg;
        return not defined explain_invalid(@_);
    }

=head3 C<explain_invalid()>

Identifies a global constraint that makes this Combination
impossible.
Returns either (in order of preference):

=over

=item * a string describing a constraint from the
constraint table, or

=item * an index into the global constraint formulas

=back

Returns C<undef> if this Combination is valid.

    if (defined(my $expl = $comb->explain_invalid())) {
        if ($expl =~ /^\d/) {
            print "Constraint '$Constraints[-1][$expl]': ",
                $Constraint_Texts[-1][$expl] . "\n";
        }
        else {
            print "Constraint table: $expl\n";
        }
    }
}

=cut

    sub explain_invalid {
        ref $_[0] or croak
            'explain_invalid()' . $_instance_err_msg;
        if (my ($if, $then) = _explain_invalid_from_table($_[0])) {
            return $if->str() . ' => ' . $then->str();
        }
        else {
            return _explain_unexpected($_[0], scalar @Events);
        }
    }

    sub _explain_invalid_from_table {
        my $comb = $_[0];
        for my $row (@_Constraint_Table) {
            my $ifvals = $$row[0];
            my $comb_val = $comb->_bits($ifvals->var_index());
            if ($comb_val & $ifvals->val_bits()) {
                for my $thenvals (map $$row[$_], 1 .. $#$row) {
                    return $ifvals, $thenvals
                        unless $comb->_bits($thenvals->var_index())
                        & $thenvals->val_bits();
                }
            }
        }
        return;
    }

=head3 C<expected()>

Returns true if the specified event (an index into @Events)
is expected in this state, otherwise false.
(An event is 'expected' in a state if the state
violates no constraints associated with that event.)

    my $count = 0;
    $state->expected($_) and ++$count for (0 .. $#Events);
    print "$count events possible in this state.\n";

=cut

    sub expected {
        ref $_[0] or croak 'expected()' . $_instance_err_msg;
        _check_event('expected', $_[1]);
        not defined _explain_unexpected(@_);
    }

    sub _check_event {
        my ($sub, $event) = @_;
        defined $event or croak $sub . '(): no event specified';
        $event =~ /^\d+$/ && $event < @Events or croak
            $sub . "(): invalid event index ($event)";
    }

=head3 C<explain_unexpected()>

Given an event (an index into ${'C<@Events>)'},
returns the index of the first event constraint
which excludes this state.
Returns C<undef> if there is no such constraint.

    if (defined (my $ci = $state->explain_unexpected($event))) {
        print "Event '$Events[$event]' is unexpected " .
            "in this state: see constraint '" .
            $Constraints[$event][$ci] . "'.\n";
    }

=cut

    sub explain_unexpected {
        my ($comb, $event) = @_;
        ref $comb or croak
            'explain_unexpected()' . $_instance_err_msg;
        _check_event('explain_unexpected', $event);
        _explain_unexpected(@_);
    }

    sub _explain_unexpected {
        my ($comb, $event) = @_;
        for my $i (0 .. $#<#---->{$_Constraint_Formulas[$event]}) {
            return $i unless ${package}::_eval_node(
                $_Constraint_Formulas[$event][$i], $comb, 1);
        }
        return;
    }

=head3 C<apply_rules()>

For a given case (i.e., Combination and event), find the
first rule that applies and get the outcome (or next state)
of that rule; returns an empty list if no rule applies.

    unless (($rule, $output) = $input->apply_rules($event)) {
        die "no rule applies for this case";
    }

If the rule allows more than one outcome (i.e., a
non-deterministic rule) then one is chosen arbitrarily.

"Extra logic" - a reference to a method - may optionally
be passed as an additional parameter:

    ($rule, $output) = $input->apply_rules($event,
        sub {
            my ($input, $event, $rule, $output) = @_;
            # return nothing if a rule is to be skipped
            return if unwanted_rule($rule);
            # an altered outcome could be returned here
            return $output;
        });

In this case, whenever a rule that applies is found,
this user-supplied method gets a chance to veto the
rule (causing C<apply_rules> to look for another rule)
or change the outcome of the rule.

=cut

    sub apply_rules {
        my ($curr_state, $event, $extra_logic) = @_;
        ref $curr_state or croak
            'apply_rules()' . $_instance_err_msg;
        _check_event('apply_rules', $event);
        if (defined $extra_logic) {
            ref($extra_logic) eq 'CODE' or croak 'apply_rules():' .
                ' extra logic is not a code reference';
        }
        for my $rule (0 .. $#<#---->{$Rules[$event]}) {
            my $table_rows = $_Rules_Table[$event][$rule];
            if ($$table_rows[0]->includes($curr_state)) {
                my $mask = $$table_rows[1]->transform($curr_state);
                my $formula = $_Rule_Formulas[$event][$rule];
                if (defined $formula) {
                    $mask = ${package}::_eval_rule_formula(
                        $formula, $curr_state, $mask);
                    next unless $mask;
                }
                my $next_state = $Combination->_from_mask($mask);
                if ($extra_logic) {
                    $next_state = &$extra_logic($curr_state,
                        $event, $rule, $next_state);
                    next unless $next_state;
                }
                return ($rule, $next_state);
            }
        }
        return;
    }

    # get the first Combination allowed by a _CombinationMask
    sub _from_mask {
        my ($class, $mask) = @_;
        my @self;
        for my $i (0 .. $#Vars) {
            if ($self[$i] = $mask->vals($i) || 0) {
                my $count = 0;
                ++$count while not $self[$i] & 1 << $count;
                $self[$i] = $count;
            }
        }
        return bless \@self, $class;
    }

=head3 C<rule_applies()>

Verify that a rule allows a particular outcome
(or, in a finite state machine model, a transition to a
particular next state).

One possible use of this method is where extra logic is
passed to C<apply_rules> to help determine the result of
a rule which allows more than one outcome:
in that case, we might want to double check that the
altered outcome returned by the extra logic remains
consistent with the rule.

    sub extra {
        # ...
    }
    ($rule, $output) = $input->apply_rules($event, \&extra);
    if (defined $rule) {
        $input->rule_applies($event, $rule, $output) or die;
    }

=cut

    sub rule_applies {
        my ($curr_state, $event, $rule, $next_state) = @_;
        ref $curr_state or croak
            'rule_applies()' . $_instance_err_msg;
        _check_event('rule_applies', $event);
        defined $rule or
            croak 'rule_applies(): no rule specified.';
        $rule =~ /^\d+$/ && $rule < @{$Rules[$event]} or croak
            "rule_applies(): " .
            "invalid rule index for event $event";
        defined $next_state or croak
            'rule_applies(): no outcome specified';
        eval { no warnings; $next_state->_bits(0); };
        croak 'rule_applies(): outcome is not a Combination' if $@;
        my $table_rows = $_Rules_Table[$event][$rule];
        return 0 unless $$table_rows[0]->includes($curr_state);
        my $mask = $$table_rows[1]->transform($curr_state);
        return 0 unless $mask->includes($next_state);
        my $formula = $_Rule_Formulas[$event][$rule];
        if (defined $formula) {
            return 0 unless ${package}::_eval_rule_formula(
                $formula, $curr_state,
                $_CombinationMask->from_comb($next_state));
        }
        return 1;
    }

=head3 C<iter()>

Returns a new iterator object with a C<next()> method
which returns each Combination in turn, beginning
with Combination number 0.

If called as an instance method, C<$c-E<gt>iter()>, then
the returned iterator starts at C<$c>.

    my $iter1 = $Combination->iter();
    while (my $c1 = $iter1->next()) {
        my $iter2 = $c1->iter();
        while (my $c2 = $iter2->next()) {
            # do something with all pairs of Combinations
            # ($c1, $c2) where $c1->num() <= $c2->num()
        }
    }

It is also possible to specify one or more variables
whose values are fixed: in this case, the iterator
skips over any Combination in which those variables
have a different value.

    # iterate over all Combinations in which the values
    # of variables 'Alpha' and 'Gamma' are 'a2' and 'g1':
    my $first = $Combination->new(Alpha=>'a2', Gamma=>'g1');
    my $iter = $first->iter('Alpha', 'Gamma');
    while (my $comb = $iter->next()) {
        # ...
    }

=cut

    sub iter {
        my @args = shift;
        for my $i (0 .. $#_) {
            defined (my $var = $_Var_Indices{$_[$i]})
                or croak "No variable '$_[$i]' in \@Vars";
            push @args, $var;
        }
        $_CombIter->new(@args);
    }

=head3 C<iter_valid()>

Returns a new iterator object with a C<next()> method
which returns each valid Combination in turn.

    my $iter = $Combination->iter_valid();
    my $count = 0;
    ++$count while defined $iter->next();
    print "The global constraints (table and formulas)\n";
    print "allow $count different Combinations.\n";

If called as an instance method, C<$c-E<gt>iter_valid()>,
then the returned iterator skips Combinations before C<$c>.

=cut

    sub iter_valid {
        my $self = shift;
        croak 'iter_valid() expects no arguments' if @_;
        my $iter = $_GangIter->new($_AllowedByTableIter->new(),
            $_AllowedByFormulasIter->new(-1));
        $iter->reset($self) if ref $self;
        $iter;
    }

=head3 C<iter_expected()>

Given an event (an index into ${'C<@Events>)'}, this
returns a new iterator object with a C<next()> method
which iterates over the states in which the event is
expected.

See the example in L</iter_gang()>.

If called as an instance method, C<$c-E<gt>iter_expected()>,
then the returned iterator skips Combinations before C<$c>.

=cut

    sub iter_expected {
        croak 'iter_expected() expects an index into @Events'
            unless defined($_[1]) && exists($Events[$_[1]]);
        my $iter = $_AllowedByFormulasIter->new($_[1]);
        $iter->reset($_[0]) if ref $_[0];
        $iter;
    }

=head3 C<iter_gang()>

Efficiently combines two or more existing iterators;
returns a new iterator which iterates over only those
Combinations which all of the specified iterators agree
should be part of the sequence.

Note: The iterators passed to C<iter_gang()>
must be independent of each other and should not
be used for anything else.

For example, the following code is likely to execute much
faster than code which tests the validity of each
"C<a2 g1>" Combination separately:

    # iterate over all valid Combinations in which the
    # values of 'Alpha' and 'Gamma' are 'a2' and 'g1':
    my $iter = $Combination->iter_gang(
        $Combination->new(Alpha=>'a2', Gamma=>'g1')
        ->iter('Alpha', 'Gamma'),
        $Combination->iter_valid());
    while (my $comb = $iter->next()) {
        # ...
    }

Similarly, this code iterates over all valid states
in which C<$event> is expected:

    my $valid_exp = $Combination->iter_gang(
        $Combination->iter_expected($event),
        $Combination->iter_valid());
    while (my $comb = $valid_exp->next()) {
        # ...
    }

In this case, however, if the code above is nested in
a loop over all the events then it may be much faster
to iterate just once over the valid states instead:

    my $iv = $Combination->iter_valid();
    while (my $state = $iv->next()) {
        for my $event (0 .. $#Events) {
            if ($state->expected($event)) {
                # ...
            }
        }
    }

=cut

    sub iter_gang {
        warnings::warnif('misc',
            'iter_gang() is a class method - invocant ignored')
            if ref shift;
        croak 'iter_gang() expects at least 2 arguments' if @_ < 2;
        (ref($_) || '') =~ /Iter$/ or croak
            'iter_gang() arguments must be iterators' for (@_);
        $_GangIter->new(@_);
    }
}

# Associates each variable with a set of allowed values.
{
    package ${package}::_CombinationMask;
    use warnings;
    use strict;

    # arguments are _Vals objects
    sub new {
        my $class = shift;
        my @self;
        for (@_) { $self[$_->var_index()] = $_->val_bits(); }
        return bless \@self, $class;
    }
    # construct a mask which allows only a specific Combination
    sub from_comb {
        my ($class, $comb) = @_;
        my @self;
        $self[$_] = $comb->_bits($_) for (0 .. $#Vars);
        return bless \@self, $class;
    }
    # is the specified Combination covered by this mask?
    sub includes {
        my ($self, $comb) = @_;
        for my $i (0 .. $#Vars) {
            my $mask = $$self[$i];
            if (defined $mask and not $mask & $comb->_bits($i)) {
                return 0;
            }
        }
        return 1;
    }
    # given pairs of (variable index, bitfield) make a mutated copy
    sub copy {
        my $self = shift;
        my @out = @$self;
        while (@_) {
            my $var_index = shift;
            $out[$var_index] = shift;
        }
        return bless \@out, ref $self;
    }
    # get, as a bitmask, the values allowed for a variable,
    # or undef if no values were specified for that variable
    sub vals {
        my ($self, $var_index) = @_;
        return $$self[$var_index];
    }
}

# The target row of a tabular transition rule.
{
    package ${package}::_TargetRow;
    use warnings;
    use strict;

    # arguments are _Vals objects; missing values for a
    # variable means 'no change'
    sub new {
        my $class = shift;
        bless [$_CombinationMask->new(@_)], $class;
    }
    # given a current state (Combination) get the allowed
    # next states; returns a _CombinationMask in which
    # values are specified for all variables
    sub transform {
        my ($self, $comb) = @_;
        my $mask = $$self[0];
        $mask->copy(map {
            defined $mask->vals($_) ? () : ($_, $comb->_bits($_)) }
            0 .. $#Vars);
    }
}

# basic sequential Combination iterator class
{
    package ${package}::_CombIter;
    use warnings;
    use strict;

    # Argument 1 (optional unless there are other arguments):
    # first Combination to return; can be $Combination or undef,
    # in which case Combination 0 is first.
    # Optional arguments: indices of variables whose values
    # are required not to change; the next() method will
    # return only Combinations in which the values of these
    # variables are the same as in the first Combination.
    sub new {
        my $self = bless { }, shift;
        my $first = shift || $Combination;
        ($self->{iter}, $self->{reset}) = $first->_iter_subs(@_);
        $self;
    }

    # Set the next Combination to return, i.e., skip back/forward.
    # Defaults to the Combination number 0.
    # However, if variables were passed to the constructor then
    # their values in the Combinations returned by the iterator
    # will continue to be the same as in the first Combination
    # that was returned by the iterator (or which would have been
    # returned if reset() had not been called).
    # (This is necessary so that it will work in a GangIter;
    # equally, for subclasses, if the argument does not satisfy a
    # condition enforced by the iterator then it should next return
    # the first Combination following it which does satisfy the
    # condition.)
    sub reset {
        my $self = shift;
        $self->{reset}->(@_);
    }

    sub next { $_[0]->{iter}->(); }
}

# Iterator over Combinations allowed by the constraint table.
{
    package ${package}::_AllowedByTableIter;
    use base '${package}::_CombIter';
    use warnings;
    use strict;

    sub next {
        my $self = $_[0];
        for (;;) {
            my $c = $self->SUPER::next();
            return unless defined $c;
            if (my ($if, $then) = $c->_explain_invalid_from_table())
            {
                my $iter = $c->_iter_change($if->var_index(),
                    $then->var_index());
                for (;;) {
                    $c = $iter->();
                    return unless defined $c;
                    last if $c->_bits($then->var_index())
                        & $then->val_bits()
                        or not ($c->_bits($if->var_index())
                        & $if->val_bits());
                }
                $self->reset($c);
            }
            else {
                return $c;
            }
        }
    }
}

# Iterator over Combinations allowed by constraint formulas.
{
    package ${package}::_AllowedByFormulasIter;
    use base '${package}::_CombIter';
    use warnings;
    use strict;

    # one argument: event index identifying the set of constraints
    sub new {
        my $self = $_[0]->SUPER::new();
        $self->{event} = $_[1];
        $self;
    }

    sub next {
        my $self = $_[0];
        my $event = $self->{event};
        for (;;) {
            my $c = $self->SUPER::next();
            return unless defined $c;
            if (defined(my $i = $c->_explain_unexpected($event))) {
                my $iter = $c->_iter_change(
                    @{$_Constraint_Vars[$event][$i]});
                for (;;) {
                    $c = $iter->();
                    return unless defined $c;
                    last if ${package}::_eval_node(
                        $_Constraint_Formulas[$event][$i], $c, 1);
                }
                $self->reset($c);
            }
            else {
                return $c;
            }
        }
    }
}

# Gang iterator - see iter_gang() for description.
{
    package ${package}::_GangIter;
    use base '${package}::_CombIter';
    use warnings;
    use strict;

    sub new { bless [@_[1 .. $#_]], $_[0]; }

    sub reset {
        my $self = shift;
        $$self[$_]->reset(@_) for 0 .. $#$self;
    }

    sub next {
        my $self = shift;
        my @comb;
        # index of mininum/maximum Combination
        my $imin = my $imax = 0;
        for my $i (0 .. $#$self) {
            defined($comb[$i] = $$self[$i]->next()) or return;
            $imax = $i if $comb[$i]->_cmp($comb[$imax]) > 0;
            $imin = $i if $comb[$i]->_cmp($comb[$imin]) < 0;
        }
        while ($comb[$imin]->_cmp($comb[$imax])) {
            $imin = $imax;
            for my $i (0 .. $#$self) {
                unless ($i == $imin) {
                    $$self[$i]->reset($comb[$imax]);
                    defined($comb[$i] = $$self[$i]->next())
                        or return;
                    $imax = $i if $comb[$i]->_cmp($comb[$imax]) > 0;
                }
            }
        }
        $comb[0];
    }
}

# equivalent to calling _eval_node() with the same parameters
# except that it catches an exception - see _eval_node()
sub _eval_rule_formula {
    my $result;
    eval { $result = _eval_node(@_); };
    if ($@) { ref $@ and return $@ or die $@; }
    return $result;
}

# Used to evaluate constraint formulas and rule formulas.
# For a constraint formula, the arguments are
#    (formula, Combination, 1)
# and 1 (the third argument) is returned if the formula
# is true for the Combination; otherwise it returns 0.
# For a rule formula, the arguments are
#    (formula, current_state, next_mask)
# where next_mask limits the possible next states that
# can be used to satisfy the formula.
# If no next state can be found to satisfy the arguments
# then this subroutine returns 0.
# Otherwise, the result is either next_mask or a
# (possibly non-unique, arbitrarily chosen) restriction
# of it, defining a set of possible next states allowed
# by the formula given the current_state.
# This result is either returned or (to allow for a neater
# implementation) thrown as an exception.
sub _eval_node {
    my ($formula, $current, $next_mask, $node) = @_;
    defined $formula or return $next_mask;
    defined $node or $node = $formula;
    $_ = $$node[0];
    if ($_ == _SOURCE) {
        my $srcvals = $$node[1];
        return $next_mask if $srcvals->val_bits()
            & $current->_bits($srcvals->var_index());
        return 0;
    }
    elsif ($_ == _TARGET) {
        # target values can occur only in rule formulas
        my $tmask = $$node[1];
        my $tmask_var = $tmask->var_index();
        my $tmask_vals = $tmask->val_bits();
        # an empty set of target values signifies 'no change'
        $tmask_vals ||= $current->_bits($tmask_var);
        my $next_vals = $next_mask->vals($tmask_var);
        my $restricted = $next_vals & $tmask_vals;
        my $excluded = $next_vals & ~$tmask_vals;
        if (not $restricted) {
            return 0;
        }
        elsif ($excluded) {
            # It is possible the overall formula will be satisfied
            # only if this clause is satisfied - or only if it is
            # not satisfied; recurse to allow for both possibilities
            my $result = _eval_node($formula, $current,
                $next_mask->copy($tmask_var, $restricted))
                || _eval_node($formula, $current,
                $next_mask->copy($tmask_var, $excluded));
            # the result would be lost at any parent _NOT nodes
            # - so smuggle it out as an exception
            die $result if $result;
            return $result;
        }
        else {
            return $next_mask
        }
    }
    elsif ($_ == _NOT) {
        _eval_node($formula, $current, $next_mask, $$node[1])
            and return 0;
        return $next_mask;
    }
    elsif ($_ == _AND) {
        for my $i (1 .. $#$node) {
            $next_mask = _eval_node($formula, $current,
                $next_mask, $$node[$i]);
            return 0 unless $next_mask;
        }
        return $next_mask;
    }
    elsif ($_ == _OR) {
        for my $i (1 .. $#$node) {
            my $new_mask = _eval_node($formula, $current,
                $next_mask, $$node[$i]);
            return $new_mask if $new_mask;
        }
        return 0;
    }
    elsif ($_ == _IMPLIES) {
        my $lhs = _eval_node($formula, $current, $next_mask,
            $$node[1]);
        return $next_mask unless $lhs;
        return _eval_node($formula, $current, $lhs, $$node[2]);
    }
    else {
        die 'Unknown parse node';
    }
}

1;

=head1 SEE ALSO

This package was generated using
FMPP, the FreeMarker PreProcessor
(L<http://fmpp.sourceforge.net/>)
and the Perl 5 code generation package for
Statestep (L<http://statestep.com>).

=cut

__END__

Note on _iter_change()
----------------------

The subroutine _iter_change() is easiest to understand
with the help of an example. Suppose:
  * there is a variable 'A' with 3 values ('a0', 'a1', 'a2')
    ('a0' being the default value) and similarly for 4 other
    variables 'B' .. 'E'
  * _iter_change() is invoked on a Combination in which
    'A' is 'a1', 'B' is 'b1', and so on.
  * the arguments passed to _iter_change() are the indices
    of variables 'A','C','D' (columns marked '**' below).
Then the closure loops as follows:

             [a1, b1, c1, d1, e1]           121
              **      **  **
count cycles      combination (c)   imax c->num
    1      3 [a1, b1, c1, d2, e0]  0 (D)    123
    2      9 [a1, b1, c2, d0, e0]  1 (C)    126
    3      9 [a1, b1, c2, d1, e0]  1 (C)    129
    4      9 [a1, b1, c2, d2, e0]  1 (C)    132
    5      9 [a1, b2, c0, d0, e0]  1 (C)    135
    6      9 [a1, b2, c0, d1, e0]  1 (C)    138
    7      9 [a1, b2, c0, d2, e0]  1 (C)    141
    8      9 [a1, b2, c1, d0, e0]  1 (C)    144
    9     27 [a2, b0, c0, d0, e0]  2 (A)    162
   10     27 [a2, b0, c0, d1, e0]  2 (A)    165
   11     27 [a2, b0, c0, d2, e0]  2 (A)    168
   12     27 [a2, b0, c1, d0, e0]  2 (A)    171
   13     27 [a2, b0, c1, d1, e0]  2 (A)    174
   14     27 [a2, b0, c1, d2, e0]  2 (A)    177
   15     27 [a2, b0, c2, d0, e0]  2 (A)    180
   16     27 [a2, b0, c2, d1, e0]  2 (A)    183
   17     27 [a2, b0, c2, d2, e0]  2 (A)    186

Note:
  * At count = 2, there is a carry to variable 'C'.
  * At count = 5, there is a carry to variable 'B' - even
    though it is not one of the fixed (**) variables.
  * At count = 9, 'A' is incremented because all possible
    combinations of variables 'C' and 'D' have been returned
    - that is, if we include the first combination (count = 0)
    among those "returned".
  * 18 out of a maximum 27 (=3*3*3) Combinations are returned
    (since no Combination with A=a0 is included).

<#--
To duplicate the above workings of _iter_change(),
add this before the line 'return sub {':

    my $format = "%5s %6s %s %6s %6s\n";
    printf $format, '', '', $self->str(), '', $self->num();
    printf $format, '', '', ' ' . join('  ', map { my $i = $_;
        grep($_==$i, @indices) ? '**' : '  ' } 0 .. $#Vars), '', '';
    print "count cycles      combination (c)   imax c->num\n";

and this before the line '++$count;':

    printf $format, $count, $cycles, $comb->str,
        "$imax ($Vars[$indices[$imax]])", $comb->num;

Then in the test/ directory do:

    make gen/fivevars_jaxen_model.pm
    perl -Igen -Mfivevars_jaxen_model=:all

and execute the following Perl code:

    $c = $Combination->new(qw/A a1 B b1 C c1 D d1 E e1/);
    $iter = $c->_iter_change(0,2,3);
    while ($iter->()) { }
-->

<#-----------------  Macros and Functions ----------------->

<#-- Output code for a set of constraint formulas. -->
<#macro constraint_formulas parent index>
    <#if parent.constraints.constraint?size == 0>
        $Constraints[${index}] = [];<#lt>
        $Constraint_Texts[${index}] = [];<#lt>
        $_Constraint_Formulas[${index}] = [];<#lt>
        $_Constraint_Vars[${index}] = [];<#lt>
        <#return>
    </#if>
    <#list parent.constraints.constraint as cf>
        $Constraints[${index}][${cf_index}] = <#t>
        '${cf.@name}';<#lt>
        $Constraint_Texts[${index}][${cf_index}] = <#t>
        ${perl_string(cf.formula.text)};<#lt>
        $_Constraint_Formulas[${index}][${cf_index}] =<#t>
        <#if cf.formula.parse_node?size != 0>
            <@parse_node_to_code node=cf.formula.parse_node/>
            ;<#lt>
        <#else>
            <#lt> undef;
        </#if>
        $_Constraint_Vars[${index}][${cf_index}] = [<#t>
        <#local var_nodes = cf[".//value_set/@variable"]/>
        <#local first_var = true/>
        <#list usable_variables as var>
            <#if var_nodes?seq_contains(var.@name)>
                <#if first_var>
                    <#local first_var = false/>
                <#else>
                    , <#t>
                </#if>
                ${var_index}<#t>
            </#if>
        </#list>
        ];<#lt>
    </#list>
</#macro>


<#-- Generate the code for a formula parse tree. -->
<#macro parse_node_to_code node indent="    ">

${indent}[<#rt>
    <#if node.@type == "values">
        <#local vals = node.value_set/>
        _${vals.@type?upper_case}, ${values_to_code(vals)}]<#t>
    <#else>
        _${node.@type?upper_case},<#t>
        <#list node.parse_node as subnode>
            <@parse_node_to_code node=subnode indent=indent + "    "/>
            <#if subnode_has_next>
                ,<#t>
            <#else>
                ]<#t>
            </#if>
        </#list>
    </#if>
</#macro>


<#--
Given a rule node, return a list of the code objects
needed to initialize the corresponding _TargetRow
object in the generated code.
-->
<#function targetvals rule>
    <#local arglist = []/>
    <#list usable_variables as var>
        <#attempt>
            <#--
            this works only if values are specified for 'var',
            otherwise the recover block is executed...
            -->
            <#local vals = rule[
            "./values_array/value_set[@type='target'][
            @variable='${var.@name}']"][0]
            />
            <#--
            no object is needed for "no change" variables
            since it is the default
            -->
            <#if vals.@content != "no_change">
                <#local arglist = arglist +
                ["${values_to_code(vals)}"]
                />
            </#if>
        <#recover>
            <#--
            if no values are listed for a variable in the
            model file the default semantic is "don't care"
            -->
            <#local arglist = arglist +
            ["$_Vals->new('${var.@name}','*')"]
            />
        </#recover>
    </#list>
    <#return arglist/>
</#function>


<#--
Convert a set of variable values to the corresponding code.
"No change" value sets produce code for an empty set of values.
-->
<#function values_to_code valset>
    <#local code = "$_Vals->new('${valset.@variable}'"/>
    <#switch valset.@content>
        <#case "no_change">
            <#break>
        <#case "dont_care">
            <#local code = code + ",'*'"/>
            <#break>
        <#case "list">
            <#local code = code + ",'"/>
            <#local children=valset.value/>
            <#if children?size == 0>
                <#-- an empty list is an implicit "don't care" -->
            <#local code = code + "*"/>
            <#else>
                <#list val_def_nodes(valset) as val>
                    <#local code = code + "${value_id(val)}"/>
                    <#if val_has_next>
                        <#local code = code + "','"/>
                    </#if>
                </#list>
            </#if>
            <#local code = code + "'"/>
            <#break>
    </#switch>
    <#return code + ")"/>
</#function>


<#--
Given a value node in a variable definition tree,
i.e., below a 'variable' element, return the string
identifying that value or group of values.
-->
<#function value_id def_node>
    <#local id = def_node.@name/>
    <#list def_node?ancestors("value") as group>
        <#local id = group.@name + "-" + id/>
    </#list>
    <#if def_node.value?size != 0>
        <#local id = id + "-*"/>
    </#if>
    <#return id/>
</#function>


<#--
Given a set of value nodes in a value set,
return the corresponding nodes in the variable
value definition tree.
-->
<#function val_def_nodes value_set>
    <#local result = []/>
    <#local varName = value_set.@variable/>
    <#local var =
    doc.model.variables["variable[@name='${varName}']"][0]
    />
    <#list value_set[".//value[not(./value)]"] as group>
        <#local def_node = var/>
        <#local path_to_group = group?ancestors("value")?reverse
        + [group]
        />
        <#list path_to_group as component>
            <#local name=component.@name/>
            <#local def_node =
            def_node["./value[@name='${name}']"][0]
            />
        </#list>
        <#local result = result + [def_node]/>
    </#list>
    <#return result>
</#function>


<#--
Convert text to a Perl string adding any necessary escapes.
-->
<#function perl_string text>
    <#return "'" +
    text?replace("\\", "\\\\")?replace("'", "\\'") + "'">
</#function>
