#!/usr/bin/env perl
# Test code for extralogic.pl.ftl
# Takes one argument: Perl code generated from that template.
# The Perl module generated from the same model using the
# main template should be specified using the -M option on
# the command line.
# This code is designed to test only the extra.s2 model.

use 5.008008;
use warnings;
use strict;

require $ARGV[0];

# this subroutine is called by the code embedded in the model
my ($allowed_event, $allowed_rule);
sub rule_allowed {
    return unless defined($allowed_event)
        && defined($allowed_rule);
    return $Events[$_[0]] eq $allowed_event
        && $Rules[$_[0]][$_[1]] eq $allowed_rule;
}

# to convert an event name to a number
my %event_index;
$event_index{$Events[$_]} = $_ for 0 .. $#Events;

# to convert a variable name to a number
my %var_index;
$var_index{$Vars[$_]} = $_ for 0 .. $#Vars;


# rule without embedded code
my $c = $Combination->new(Weather=>'snow', Wind=>'calm');
my $event = $event_index{'decide'};
my ($rule, $out) = $c->apply_rules($event, \&extralogic);
my $activity = $var_index{'Activity'};
die unless $Rules[$event][$rule] eq 'A';
die unless $Values[$activity][$out->value($activity)] eq 'ski';

# combination for which there is no rule
$c = $Combination->new(Weather=>'fog', Wind=>'calm');
($rule, $out) = $c->apply_rules($event, \&extralogic);
die if defined $rule or defined $out;

# (for event 'decide', this is covered by rules 'A_1' and 'A_2')
$c = $Combination->new(Temperature=>'hot',
    Weather=>'fine', Wind=>'breeze');

my $extra2 = sub { print "== @_[0..1] ", $_[2]->str(), " ==\n"; my $x = extralogic(@_);
print $x->str(), "\n" if defined $x; return $x };

# all applicable rules vetoed
($rule, $out) = $c->apply_rules($event, \&extralogic);
die if defined $rule or defined $out;

# event is checked as well as rule (veto for wrong event)
$allowed_event = 'event2';
$allowed_rule = 'A_1';
($rule, $out) = $c->apply_rules($event, \&extralogic);
die if defined $rule or defined $out;

# first potentially vetoable rule not vetoed
$allowed_event = 'decide';
($rule, $out) = $c->apply_rules($event, \&extralogic);
die unless $Rules[$event][$rule] eq 'A_1';
die unless $Values[$activity][$out->value($activity)] eq 'windsurf';

# first vetoable rule vetoed, later rule not vetoed
$allowed_rule = 'A_2';
($rule, $out) = $c->apply_rules($event, \&extralogic);
die unless $Rules[$event][$rule] eq 'A_2';
die unless $Values[$activity][$out->value($activity)] eq 'hillwalk';

# conditional changing of variable values (a)
$allowed_event = $allowed_rule = undef;
$event = $event_index{'event2'};
($rule, $out) = $c->apply_rules($event, \&extralogic);
die unless $Rules[$event][$rule] eq 'A';
die unless $Values[$activity][$out->value($activity)] eq 'hillwalk';

# conditional changing of variable values (b)
$allowed_event = 'event2';
$allowed_rule = 'A';
($rule, $out) = $c->apply_rules($event, \&extralogic);
die unless $Rules[$event][$rule] eq 'A';
die unless $Values[$activity][$out->value($activity)] eq 'windsurf';
