#!/usr/bin/env perl
# Standard test for code generated from module.pm.ftl.
# Assumes the module being tested is imported
# (use the -M option on the command line).

use 5.008008;
use warnings;
use strict;

# for bottom-up testing, we access some of the
# hidden symbols in the generated code
our ($pack, @_Constraint_Formulas);
BEGIN {
    ($pack = $Combination) =~ s/::Combination$//;
    for (qw/_Constraint_Formulas _eval_node/) {
        no strict 'refs';
        *$_ = \*{$pack . "::$_"};
    }
}

# limited tests for big model
if ($Combinations > Math::BigInt->new('2')->bpow('32')) {
    # test handling of numbers outside usual integer range
    print "Combinations: $Combinations\n";
    my $last = $Combination->num($Combinations - 1);
    $last->num() == $Combinations - 1
        or die "Combination->num() 1";
    $last = $Combination->new(
        map { $Vars[$_], $Values[$_][-1] } 0 .. $#Vars);
    $last->num() == $Combinations - 1
        or die "Combination->num() 2";
    # this iteration is only realistic if the large model is
    # engineered to have a very small number of valid states
    print "Counting valid combinations ...\n";
    my $vi = $Combination->iter_valid();
    my $count = 0;
    ++$count while $vi->next();
    print "Valid combinations: $count\n";
    # other tests would be too slow and produce huge outputs
    exit 0;
}

# output variables and their values
print "Variables:\n";
for my $var (0 .. $#Vars) {
    no warnings 'once';
    for my $val (@{$Values[$var]}) {
        print "$Vars[$var] $val\n";
    }
}

# indexed by combination number; true iff combination valid
my @valid;

print "\nConstraint Table disallows:\n";
my $iter = $Combination->iter();
# test $Combination->num() with Math::BigInt;
# test iterator $Combination->iter() without arguments
my $prev;
my $iter2;
my $half = int($Combinations / 2);
for (my $n = Math::BigInt->bzero(); $n < $Combinations; ++$n) {
    my $copy = $n->copy();
    my $c = $Combination->num($n) or
        die 'all combinations should be boolean true (overload)';
    $n->bcmp($copy) and die 'argument to num() mutated';
    $c->num()->bcmp($n) and die '$Combination->num() (BigInt)';
    if (my ($if, $then) = $c->_explain_invalid_from_table()) {
        $valid[$n] = 0;
        print $c->str, "\n";
        # test explain_invalid() returning constraint from table
        my $exp = $if->str() . ' => ' . $then->str();
        my $out = $c->explain_invalid();
        $out eq $exp or die "expected '$exp', got '$out'";
    }
    else {
        $valid[$n] = 1;
    }
    # test $Combination->_cmp()
    # test overloading of comparison operators
    $c->_cmp($c) and die;
    $c == $Combination->num($n) or die;
    $c eq $Combination->num($n) or die;
    if (defined $prev) {
        $c->_cmp($prev) > 0 or die;
        $prev->_cmp($c) < 0 or die;
        $c >= $prev && $c > $prev && $c != $prev or die;
        $prev <= $c && $prev < $c && !($prev == $c) or die;
        ($c cmp $prev) > 0 && ($c <=> $prev) > 0 or die;
        ($prev cmp $c) < 0 && ($prev <=> $c) < 0 or die;
        $prev le $c && $prev lt $c or die;
        $c ge $prev && $c gt $prev && $c ne $prev or die;
    }
    $prev = $c;
    # test $Combination->new() with initializers for all/some vars
    # test $Combination->new() with no arguments
    # test combination->new() (i.e., called on an instance)
    # test $Combination->value() in numeric mode
    # test $Combination->value() in string mode
    my (@init, @reset);
    for my $var (0 .. $#Vars) {
        my $val = $c->value($var);
        $c->value($Vars[$var]) eq $Values[$var][$val] or die;
        if ($val) {
            # omit default values to get full/partial/empty list
            push @init, $Vars[$var], $Values[$var][$val];
            # minimal initializers to get back default comb.
            push @reset, $Vars[$var], $Values[$var][0];
        }
    }
    my $c2 = $Combination->new(@init);
    my $cdef = $c2->new(@reset);
    my $c3 = $c2->new(@init);
    my $c4 = $c2->new();
    # (also tests that x->new() does not change x)
    $c2->num()->bcmp($n) and die '$Combination->new(@init)';
    $cdef->num() and die 'combination->new(@reset)';
    $c3->num()->bcmp($n) and die 'combination->new(@init)';
    $c4->num()->bcmp($n) and die 'combination->new()';
    # test iterator, normal case
    $iter->next()->num() == $n or die;
    # test independence of iterators
    if (!defined($iter2) and $n == $half) {
        $iter2 = $Combination->iter();
        # test iter() as instance method
        $iter = $c->iter();
        $iter->next()->num() == $n or die;
    }
    if (defined $iter2) {
        $iter2->next()->num() + $half == $n or die;
    }
}
# test iterator exhausted
defined $iter->next() and die for (0..1);
# make sure we executed the 'halfway' test code above
defined $iter2 or die;

# test sort (requires overloaded 'cmp' operator)
my @unsort = map $Combination->num($_), $Combinations - 1, 0;
push @unsort, undef;
push @unsort, reverse @unsort;
{
    # sort warns on comparisons with undef
    # (so does $Combination->_cmp - tested later)
    no warnings 'uninitialized';
    my @sort = sort @unsort;
    for (1 .. $#sort) {
        my ($prev, $curr) = @sort[$_-1,$_];
        ($prev && $prev->num()) <= ($curr && $curr->num()) or die;
    }
}

# test _AllowedByTableIter
my $abti = "${pack}::_AllowedByTableIter"->new();
for (0 .. 1) {
    for my $i (0 .. $#valid) {
        if ($valid[$i]) {
            my $c = $abti->next();
            $c->num() == $i or die;
        }
    }
    defined $abti->next() and die for (0 .. 1);
    $abti->reset();
}

test_gang_iter();

# true for indexes of combinations allowed by formulas
my @validf = (1) x $Combinations;

# test $Combination->num() with scalar
for my $i (0 .. $#{$Constraints[-1]}) {
    my $name = $Constraints[-1][$i];
    print "\nGlobal Constraint '$name' disallows:\n";
    for my $n (0 .. $Combinations - 1) {
        my $c = $Combination->num($n);
        $c->num()->bcmp($n) and die '$Combination->num() (scalar)';
        if (!_eval_node($_Constraint_Formulas[-1][$i], $c, 1)) {
            print $c->str(), "\n";
            $validf[$n] = 0;
            if ($valid[$n]) {
                $valid[$n] = 0;
                # test explain_invalid() returning a formula index
                # (it should return the same constraint as it
                # is the first one making this state invalid)
                my $out = $c->explain_invalid();
                $out eq $i or die "expected '$name', got '$out'";
            }
        }
    }
}

# test _AllowedByFormulasIter for global constraints
my $abfi = "${pack}::_AllowedByFormulasIter"->new(-1);
for (0 .. 1) {
    for my $i (0 .. $#validf) {
        if ($validf[$i]) {
            my $c = $abfi->next();
            $c->num() == $i or die;
        }
    }
    defined $abfi->next() and die for (0 .. 1);
    $abfi->reset();
}

# test valid()
# test explain_invalid() returning undef
# test iter_valid()
my $iter_valid = $Combination->iter_valid();
my $exp_next;
for my $n (0 .. $Combinations - 1) {
    my $c = $Combination->num($n);
    $c->valid() == $valid[$n] or die "expected $valid[$n] for $n";
    # test iter_valid() as instance method
    my $iv2 = $c->iter_valid();
    if ($valid[$n]) {
        ! defined $c->explain_invalid() or die;
        my $c = $iter_valid->next();
        $c->num() == $n or die;
        die if $n && $c->_cmp($exp_next);
        $exp_next = $iv2->next();
        die if $c->_cmp($exp_next);
        $exp_next = $iv2->next();
        die if defined($exp_next) && !$c->_cmp($exp_next);
    }
    else {
        my $cn = $iv2->next();
        die if defined($exp_next) && $cn->_cmp($exp_next);
    }
}
defined $iter_valid->next() and die for (0 .. 1);
die if defined $exp_next;

test_iter_args();

for my $event (0 .. $#Events) {
    my @unexpected;
    for my $i (0 .. $#{$Constraints[$event]}) {
        print "\nEvent '", $Events[$event], "', Constraint '",
            $Constraints[$event][$i], "' disallows:\n";
        for my $n (0 .. $Combinations - 1) {
            my $c = $Combination->num($n);
            if (!_eval_node($_Constraint_Formulas[$event][$i],
                $c, 1))
            {
                print $c->str(), "\n";
                if (!$unexpected[$n]) {
                    $unexpected[$n] = 1;
                    # test explain_unexpected() (unexpected event)
                    # (it should return the same constraint as it
                    # is the first one making the event unexpected)
                    my $out = $c->explain_unexpected($event);
                    $out == $i or die "expected $i, got $out";
                }
            }
        }
    }
    # test expected()
    # test explain_unexpected() (expected event)
    # test iter_expected() alone and ganged with valid iter
    # (latter is a test of a gang iter in a gang iter)
    my $iter_exp = $Combination->iter_expected($event);
    my $iter_valid_exp = $Combination->iter_gang(
        $Combination->iter_expected($event),
        $Combination->iter_valid());
    my $next_e;
    for my $n (0 .. $Combinations - 1) {
        my $c = $Combination->num($n);
        my $expected = $c->expected($event);
        $unexpected[$n] ? not $expected : $expected
            or die "expected($event) for combination $n";
        # test iter_expected() as instance method
        my $iter2 = $c->iter_expected($event);
        if ($expected) {
            die if defined $c->explain_unexpected($event);
            my $c2 = $iter_exp->next();
            $c2->num() == $n or die;
            die if $n && $c->_cmp($next_e);
            $next_e = $iter2->next();
            die if $c->_cmp($next_e);
            $next_e = $iter2->next();
            die if defined($next_e) && !$c->_cmp($next_e);
            if ($valid[$n]) {
                my $c2 = $iter_valid_exp->next();
                $c2->num() == $n or die;
            }
        }
        else {
            my $cn = $iter2->next();
            die if defined($next_e) && $cn->_cmp($next_e);
        }
    }
    defined $iter_exp->next() and die for (0 .. 1);
    die if defined $next_e;
    defined $iter_valid_exp->next() and die for (0 .. 1);
    my @covered;
    for my $i (0 .. $#{$Rules[$event]}) {
        print "\nEvent '", $Events[$event], "', Rule '",
            $Rules[$event][$i], "' transitions:\n";
        my $extra_logic = sub {
                my ($curr, $ev, $rule, $next) = @_;
                unless ($covered[$curr->num()]++) {
                    # test apply_rules() without extralogic
                    my ($r2, $n2) = $curr->apply_rules($event);
                    die 'apply_rules() w/o extralogic'
                        unless $r2 == $rule
                        && $n2->num() == $next->num();
                }
                die 'extra_logic() gets wrong event'
                    unless $ev == $event;
                return $rule == $i ? $next : undef;
            };
        for my $c (0 .. $Combinations - 1) {
            my $current = $Combination->num($c);
            # test apply_rules() with extralogic, no change
            my ($rule, $next_state) =
                $current->apply_rules($event, $extra_logic);
            if (defined $rule) {
                die "extra_logic failed" unless $rule == $i;
                # test rule_applies() succeeds if apply_rules() does
                die "apply_rules() contradicts rule_applies()"
                    unless $current->rule_applies($event,
                    $rule, $next_state);
                # test apply_rules() succeeds when it should
                print $current->str(), "\n";
                # test apply_rules() - extralogic changes state
                my $changed = $Combination->num(
                    ($next_state->num() + 1) % $Combinations);
                my (undef, $out) = $current->apply_rules($event,
                    sub {
                        my ($curr, $ev, $rule, $next) = @_;
                        return unless $rule == $i;
                        die 'extra_logic() gets wrong next state'
                            unless $next->num()
                            == $next_state->num();
                        return $changed;
                    });
                die 'extra_logic() return value lost'
                    unless $out == $changed;
            }
            else {
                die "next_state returned without rule"
                    if defined $next_state;
            }
            for my $n (0 .. $Combinations - 1) {
                my $next = $Combination->num($n);
                # test apply_rules() succeeds if rule_applies() does
                if ($current->rule_applies($event, $i, $next)) {
                    die 'rule_applies() contradicts apply_rules()'
                        unless defined $rule;
                    # test rule_applies()
                    print '->', $next->str(), "\n";
                }
            }
        }
    }
}

# test $Combination->new() error cases
for my $c ($Combination, $Combination->new()) {
    eval { $c->new(0); };
    die unless $@ =~/variable/;
    eval { $c->new('err'); };
    die unless $@ =~/variable/;
    eval { $c->new($Vars[0]); };
    die unless $@ =~/value/;
    eval { $c->new($Vars[0], 0); };
    die unless $@ =~/value/;
    eval { $c->new($Vars[0], 'err'); };
    die unless $@ =~/value/;
}

# test $Combination->num() error cases
eval { $Combination->num(); };
die unless $@ =~/argument/;
eval { $Combination->num($Combinations); };
die unless $@ =~/range/;
eval { $Combination->num(-1); };
die unless $@ =~/positive/;
eval { $Combination->num([]); };
die unless $@ =~/integer/;
eval { $Combination->new()->num(4); };
die unless $@ =~/argument/;

# test $Combination->str() error cases
eval { $Combination->str(); };
die unless $@ =~/instance/;

# test $Combination->value() error cases
eval { $Combination->value(0); };
die unless $@ =~/instance/;
eval { $Combination->new()->value(); };
die unless $@ =~/argument/;
eval { $Combination->new()->value($#Vars + 1); };
die unless $@ =~/value()/;
eval { $Combination->new()->value(-1); };
die unless $@;
eval { $Combination->new()->value('no_such_var'); };
die unless $@;

# test overloaded comparison operator error cases
eval { my $x = $Combination->new() <= 42; };
die unless $@ =~/compared/;
eval { my $x = $Combination->new() eq 'string'; };
die unless $@ =~/compared/;
eval { my $x = $Combination->new() > {}; };
die unless $@ =~/compared/;
eval { my $x = $Combination->new() <=> []; };
die unless $@ =~/compared/;
# test overloaded comparison operators with undef
{
    use warnings;
    my $wmsg;
    local $SIG{'__WARN__'} = sub { $wmsg = $_[0] };
    $Combination->new() == undef and die;
    die unless $wmsg && $wmsg =~ /uninitialized/;
    $wmsg = undef;
    $wmsg < $Combination->new() or die;
    die unless $wmsg && $wmsg =~ /uninitialized/;
    $wmsg = undef;
    ($Combination->new() cmp undef) > 0 or die;
    die unless $wmsg && $wmsg =~ /uninitialized/;
}

# test valid() error cases
eval { $Combination->valid(); };
die unless $@ =~ /instance/;

# test explain_invalid() error cases
eval { $Combination->explain_invalid(); };
die unless $@ =~ /instance/;

# test expected() error cases
eval { $Combination->expected(); };
die unless $@ =~ /instance/;
eval { $Combination->new()->expected(0 + @Events); };
die unless $@ =~ /event/;

# test explain_unexpected() error cases
eval { $Combination->explain_unexpected(); };
die unless $@ =~ /instance/;
eval { $Combination->new()->explain_unexpected('name'); };
die unless $@ =~ /event/;

# test apply_rules() error cases
eval { $Combination->apply_rules(); };
die unless $@ =~ /instance/;
eval { $Combination->new()->apply_rules(-1); };
die unless $@ =~ /event/;
if (@Events) {
    eval { $Combination->new()->apply_rules(0, []); };
    die unless $@ =~ /code reference/;
}

# test rule_applies() error cases
eval { $Combination->rule_applies(); };
die unless $@ =~ /instance/;
eval { $Combination->new()->rule_applies(); };
die unless $@ =~ /event/;
if (@Events) {
    my $c = $Combination->new();
    eval { $c->rule_applies(0); };
    die unless $@ =~ /rule/;
    eval { $c->rule_applies(0, 'name', $c); };
    die unless $@ =~ /rule/;
    eval { $c->rule_applies(0, -1, $c); };
    die unless $@ =~ /rule/;
    for my $event (0 .. $#Events) {
        my $count = @{$Rules[$event]};
        eval { $c->rule_applies($event, $count, $c) };
        die unless $@ =~ /rule/;
        if ($count) {
            eval { $c->rule_applies($event, $count - 1); };
            die unless $@ =~ /outcome/;
            eval { $c->rule_applies($event, $count - 1, []); };
            die unless $@ =~ /outcome/;
        }
    }
}

# test iter() error cases
eval { $Combination->iter('err'); };
die unless $@ =~ /variable 'err'/;

# test iter_valid() error cases
eval { $Combination->iter_valid(0); };
die unless $@ =~ /expects/;

# test iter_expected() error cases
eval { $Combination->iter_expected(); };
die unless $@ =~ /expects/;

###################### subroutines ######################

# Test $Combionation->iter_gang()
{
    package CondIter;
    no strict;
    use base "${pack}::_CombIter";
    sub new {
        my $self = shift()->SUPER::new();
        $self->{condition} = shift();
        $self;
    }
    sub next {
        my $self = shift();
        while (defined($c = $self->SUPER::next())) {
            return $c if $self->{condition}->($c);
        }
        return;
    }
}
sub test_gang_iter {
    # Use a range of offsets to ensure we cover both the case
    # where the first combination (combination 0) and the case
    # where the last combination are among those iterated over.
    for my $offset (0 .. 9) {
        my $mod2 = CondIter->new(
            sub {not(($_[0]->num() + $offset) % 2)});
        my $mod2b = CondIter->new(
            sub {not(($_[0]->num() + $offset) % 2)});
        my $mod5 = CondIter->new(
            sub {not(($_[0]->num() + $offset) % 5)});
        my $mod10 = CondIter->new(
            sub {not(($_[0]->num() + $offset) % 10)});
        for my $iters ([$mod2, $mod5, $mod10],
            [$mod10, $mod5, $mod2],
            [$mod5, $mod2, $mod2b],
            [$mod2, $mod5, $mod2b],
            [$mod2, $mod2b, $mod10])
        {
            my $gang = $Combination->iter_gang(@$iters);
            my @range = $Combinations > 20
                ? (0, 7, $Combinations - 10, $Combinations - 1)
                : (0, $Combinations - 1);
            for my $start (@range) {
                $gang->reset($Combination->num($start));
                ++$start while ($start + $offset) % 10;
                my $exp = $start;
                while ($exp < $Combinations) {
                    my $c = $gang->next();
                    $c->num() == $exp or die $exp;
                    $exp += 10;
                }
                defined $gang->next() and die for (0 .. 1);
                $gang->reset();
                my $c = $gang->next();
                $exp = -$offset;
                $exp += 10 if $exp;
                if ($exp < $Combinations) {
                    $c->num() == $exp or die;
                }
                else {
                    die if defined $c;
                }
            }
            # (reset constituent iterators for re-use)
            $gang->reset();
        }
        eval { $Combination->iter_gang($mod2); };
        die unless $@ =~ /2 arguments/;
        eval { $Combination->iter_gang($mod2, $Combination->new); };
        die unless $@ =~ /iterator/;
        eval { $Combination->iter_gang(42, $mod2); };
        die unless $@ =~ /iterator/;
        use warnings;
        my $wmsg = '';
        local $SIG{'__WARN__'} = sub { $wmsg = $_[0] };
        $Combination->new()->iter_gang($mod2, $mod5);
        $wmsg =~ /class method/ or die "warning '$wmsg'";
        no warnings 'misc';
        $wmsg = undef;
        $Combination->new()->iter_gang($mod2, $mod5);
        defined $wmsg and die "warning '$wmsg'";
    }
}

# test $Combination->iter() with arguments - individually and
# together with an iter_valid() in a gang iterator
sub test_iter_args {
    my @var_sets;
    push @var_sets, @{subsets([0 .. $#Vars], $_)} for (1 .. @Vars);
    my @iters;
    my @var_vals;
    my $comb = $Combination;
    for my $vsi (0 .. $#var_sets) {
        my @var_set = @{$var_sets[$vsi]};
        @var_set = reverse @var_set unless $vsi % 2;
        my %var_val;
        for (0 .. $#var_set) {
            $var_val{$var_set[$_]} =
                (1 + $vsi + $_) % @{$Values[$var_set[$_]]};
        }
        push @var_vals, \%var_val;
        $comb = $comb->new(
            map {$Vars[$_], $Values[$_][$var_val{$_}]}
            keys %var_val);
        my $iter = $comb->iter(map $Vars[$_], @var_set);
        my $c = $iter->next();
        $c->_cmp($comb) and die $c->str . " " . $comb->str;
        $iter->reset();
        push @iters, $iter;
    }
    my @prev;
    my $base_iter = $Combination->iter();
    while (my $c = $base_iter->next()) {
ITER:
        for my $vsi (0 .. $#var_sets) {
            for my $var (keys %{$var_vals[$vsi]}) {
                next ITER
                    if $c->value($var) != $var_vals[$vsi]{$var};
            }
            my $iterated = $iters[$vsi]->next();
            if ($iterated->_cmp($c)) {
                die $iterated->str . " != " . $c->str;
            }
            # test reset()s to intermediate (non-iterated) combs
            my $iter2 = $iterated->iter(
                map $Vars[$_], @{$var_sets[$vsi]});
            my $previous = $prev[$vsi];
            if (!defined $previous) {
                $previous = $Combination->new();
                next ITER unless $previous->_cmp($iterated);
            }
            my $base2 = $previous->iter();
            $base2->next();
            for (;;) {
                my $non_iter = $base2->next();
                $iter2->reset($non_iter);
                my $follow = $iter2->next();
                if ($follow->_cmp($iterated)) {
                    die $follow->str . " != " . $iterated->str;
                }
                last unless $non_iter->_cmp($iterated);
            }
            $prev[$vsi] = $iterated;
        }
    }
    for my $vsi (0 .. $#var_sets) {
        defined $iters[$vsi]->next() and die for (0..1);
        my $last = $prev[$vsi] || $Combination->new();
        my $base = $last->iter();
        $base->next();
        while (defined (my $non_iter = $base->next())) {
            $iters[$vsi]->reset($non_iter);
            my $follow = $iters[$vsi]->next();
            die $follow->str() if defined $follow;
        }
    }
    # test in a gang iter;
    # test a $Combination->iter_valid() in a gang iter
    for my $vsi (0 .. $#var_sets) {
        my $c = $Combination->new(
            map {$Vars[$_], $Values[$_][$var_vals[$vsi]{$_}]}
            keys %{$var_vals[$vsi]});
        $iters[$vsi] = "${pack}::_GangIter"->new(
            $c->iter(map $Vars[$_], keys %{$var_vals[$vsi]}),
            $Combination->iter_valid());
    }
    $base_iter = $Combination->iter_valid();
    while (my $c = $base_iter->next()) {
ITER:
        for my $vsi (0 .. $#var_sets) {
            for my $var (keys %{$var_vals[$vsi]}) {
                next ITER
                    if $c->value($var) != $var_vals[$vsi]{$var};
            }
            my $iterated = $iters[$vsi]->next();
            if ($iterated->_cmp($c)) {
                die $iterated->str . " != " . $c->str;
            }
        }
    }
    for my $vsi (0 .. $#var_sets) {
        defined $iters[$vsi]->next() and die for (0..1);
    }
}

# Returns a reference to an array of all subsets from the
# specified list, each subset being a reference to an
# array listing its elements in the same order as in the
# original array.
# Arg 1: ref to an array from which the subsets are to be taken
# Arg 2: size (number of elements) in each subset;
sub subsets {
    my ($set, $n) = @_;
    my @subsets;
    if ($n) {
        for my $i (0 .. @$set - $n) {
            my $subs = subsets([@$set[$i + 1..$#$set]], $n - 1);
            for my $sub (@$subs) {
                push @subsets, [$$set[$i], @$sub];
            }
        }
    }
    else {
        push @subsets, [];
    }
    [@subsets];
}
